
import json
import argparse
import os
from pathlib import Path
import time
import logging
from kubernetes import client, config

from azure.identity import DeviceCodeCredential, DefaultAzureCredential
from azure.core.exceptions import ClientAuthenticationError
from azure.mgmt.resource.subscriptions import SubscriptionClient
from azure.mgmt.resource import ResourceManagementClient
from azure.mgmt.compute import ComputeManagementClient

logger = logging.getLogger(__name__)

def get_storage_account_name_from_disk_id(disk_id):
    disk_id_parts = disk_id.split("/")
    subscription = disk_id_parts[2]
    resource_group = disk_id_parts[4]
    storage_acount_name = disk_id_parts[8]
    return {"subscription": subscription, "resource_group": resource_group, "storage_acount_name": storage_acount_name}

def get_azure_disk_info_from_disk_uri(credential, disk_uri):
    disk_uri_parts = disk_uri.split("/")
    subscription = disk_uri_parts[2]
    resource_group = disk_uri_parts[4]
    disk_name = disk_uri_parts[8]

    compute_client = ComputeManagementClient(credential, subscription)
    disk = compute_client.disks.get(resource_group, disk_name)
    return disk


def get_all_used_pvcs(k8s_client):

    logger.debug("Fetching all pods")
    pods = k8s_client.list_pod_for_all_namespaces(watch=False)
    if pods is None:
        logger.error("No pods found")
        exit(1)

    logger.debug("Found %d pods" % len(pods.items))

    used_pvcs = set()
    for pod in pods.items:
        if pod.spec.volumes is None:
            continue
        for volume in pod.spec.volumes:
            if volume.persistent_volume_claim is not None:
                used_pvcs.add(volume.persistent_volume_claim.claim_name)

    logger.debug("Returning %d used PVCs" % len(used_pvcs))

    return used_pvcs

def remove_unnecessary_parameters_from_pv(persistent_volume):
    persistent_volume.status = None
    persistent_volume.metadata.resource_version = None
    persistent_volume.metadata.creation_timestamp = None
    persistent_volume.metadata.generation = None
    persistent_volume.metadata.managed_fields = None
    persistent_volume.metadata.uid = None
    persistent_volume.metadata.self_link = None
    persistent_volume.metadata.finalizers = None
    return persistent_volume

def remove_unnecessary_parameters_from_pvc(persistent_volume_claim):
    persistent_volume_claim.status = None
    persistent_volume_claim.metadata.resource_version = None
    persistent_volume_claim.metadata.creation_timestamp = None
    persistent_volume_claim.metadata.generation = None
    persistent_volume_claim.metadata.managed_fields = None
    persistent_volume_claim.metadata.uid = None
    persistent_volume_claim.metadata.annotations = None
    return persistent_volume_claim


def main(args):

    # Create output directory
    Path(args.logs_directory).mkdir(parents=True, exist_ok=True)
    log_file_path = os.path.join(args.logs_directory, f"{args.context}-{args.storage_class_name}-log.txt")

    # Configure logging
    logger.setLevel(logging.DEBUG)
    # Create handlers
    c_handler = logging.StreamHandler()
    c_handler.setLevel(level=args.loglevel)
    # Create formatters and add it to handlers
    c_format = logging.Formatter("[%(asctime)s] - [%(levelname)s]: %(message)s")
    c_handler.setFormatter(c_format)
    # Add handlers to the logger
    logger.addHandler(c_handler)

     # Create file handler
    f_handler = logging.FileHandler(log_file_path)
    f_handler.setLevel(level=args.loglevel)
    # Add formatter to file handler
    f_handler.setFormatter(c_format)
    # Add file handler to the logger
    logger.addHandler(f_handler)

    logger.debug("Logging configured")
    logger.debug("Using logs directory: %s" % args.logs_directory)

    # Create kubernetes clients
    config.load_kube_config()
    api_client=config.new_client_from_config(context=args.context)
    k8s_storageclass_client = client.StorageV1Api(api_client)
    k8s_client = client.CoreV1Api(api_client)

    logger.debug("Kubernetes clients created")

    # Create Azure credentials
    credential = None
    try:
        credential = DefaultAzureCredential(
            exclude_environment_credential=True)
        token = credential.get_token("https://management.core.windows.net/.default")
    except ClientAuthenticationError:
        credential = DeviceCodeCredential()

    logger.debug("Azure credentials created")

    logger.debug("Fetching used PVCs")
    used_pvcs = get_all_used_pvcs(k8s_client)


    logger.debug("Fetching storage class %s" % args.storage_class_name)
    storage_class = None
    try:
        storage_class = k8s_storageclass_client.read_storage_class(args.storage_class_name)
    except Exception as e:
        logger.error("Error reading storage class %s. Storage class: %s doesn't exist." % (args.storage_class_name))
        exit(1)



    logger.debug("Getting skuName from storage class %s" % args.storage_class_name)
    sku_name = None

    for key in storage_class.parameters.keys():
        if key.lower() == "skuname":
            sku_name = storage_class.parameters[key]
            logger.debug("Bingo, found skuName parameter")
            break   
        if key.lower() == "storageaccounttype":
            sku_name = storage_class.parameters[key]
            logger.debug("Bingo, found storageAccountType parameter")
            break
    if sku_name is None:
        logger.debug("Storage class does not have skuName or storageAccountType parameter. Using default value \"Standard_LRS\"")
        sku_name = "Standard_LRS"

    logger.debug("Using skuName: `%s`" % sku_name)
    
    logger.debug("Fetching all PVs")
    pvs = k8s_client.list_persistent_volume(watch=False)
    if pvs is None:
        logger.error("No PVs found")
        exit(1)
    logger.debug("Found %d PVs" % len(pvs.items))
    for pv in pvs.items:
        if pv.spec.storage_class_name == args.storage_class_name and pv.spec.csi is None and pv.spec.claim_ref.name not in used_pvcs:
            pvc = k8s_client.read_namespaced_persistent_volume_claim(pv.spec.claim_ref.name, pv.spec.claim_ref.namespace)
            logger.debug(json.dumps(pvc.to_dict(), indent=4, default=str))
            if pv.spec.azure_disk is not None:
                logger.debug("Fetching disk info from disk uri %s"%pv.spec.azure_disk.disk_uri)
                disk = get_azure_disk_info_from_disk_uri(credential, pv.spec.azure_disk.disk_uri)
                if disk is not None:
                    requested_size_gib = disk.disk_size_gb
                else:
                    if pv.status.capacity is not None and pv.status.capacity["storage"] is not None:
                        if pv.status.capacity["storage"].endswith.lower().endswith("ti"):
                            requested_size_gib = int(pv.status.capacity["storage"][:-2]) * 1024
                        else:
                            requested_size_gib = int(pv.status.capacity["storage"][:-2])
                sku_name = disk.sku.name
                if pv.kind is None:
                    logger.debug("Setting kind to PersistentVolume")
                    pv.kind = "PersistentVolume"
                if pv.api_version is None:
                    logger.debug("Setting apiVersion to v1")
                    pv.api_version = "v1"
                pv.spec.csi = {
                    "driver": "disk.csi.azure.com",
                    "volumeHandle": pv.spec.azure_disk.disk_uri,
                    "volumeAttributes": {
                        "csi.storage.k8s.io/pv/name": pv.spec.azure_disk.disk_name,
                        "csi.storage.k8s.io/pvc/name": pv.spec.claim_ref.name,
                        "csi.storage.k8s.io/pvc/namespace": pv.spec.claim_ref.namespace,
                        "kind": pv.spec.azure_disk.kind,
                        "requestedsizegib": str(requested_size_gib),
                        #"storage.kubernetes.io/csiProvisionerIdentity": "16XXXXXX-XXXXX-disk.csi.azure.com",
                        "storageaccounttype": "Premium_LRS" if args.storage_class_name == "managed-premium" else "Standard_LRS"
                    }
                }
                logger.debug("CSI specification added and removing azureDisk from PV")
                pv.spec.azure_disk = None
            elif pv.spec.azure_file is not None:
                resource_groupname = storage_class.parameters["resourceGroup"] if "resourceGroup" in storage_class.parameters.keys() else "NOTSET"
                account_name = storage_class.parameters["storageAccount"] if "storageAccount" in storage_class.parameters.keys() else ""
                file_share_name = pv.spec.azure_file.share_name
                secret_namespace = pv.spec.azure_file.secret_namespace
                pv.spec.csi = {
                    "driver": "file.csi.azure.com",
                    # {resource-group-name}#{account-name}#{file-share-name}#{placeholder}#{uuid}#{secret-namespace}
                    "volumeHandle": resource_groupname + "#" + account_name + "#" + file_share_name + "#" + "" + "#" + "" + "#" + secret_namespace,
                    "volumeAttributes": {
                        "shareName": pv.spec.azure_file.share_name
                    },
                    "nodeStageSecretRef": {
                        "name": pv.spec.azure_file.secret_name,
                        "namespace": pv.spec.azure_file.secret_namespace
                    }
                }
                logger.debug("CSI specification added and removing azureFile from PV")
                pv.spec.azure_file = None
            logger.debug("Check if we need to set the reclaim policy to Retain")
            if pv.spec.persistent_volume_reclaim_policy != "Retain":
                k8s_client.patch_persistent_volume(pv.metadata.name, {"spec": {"persistentVolumeReclaimPolicy": "Retain"}})
                logger.debug("Reclaim policy set to Retain")
           
            logger.debug("Removing unnecessary parameters from PV and PVC")
            
            # pvc_sanitized = remove_unnecessary_parameters_from_pvc(pvc)
            pv_sanitized = remove_unnecessary_parameters_from_pv(pv)
            logger.debug("Deleting the  PV %s" % pv.metadata.name)
            # k8s_client.delete_namespaced_persistent_volume_claim(pv.spec.claim_ref.name, pv.spec.claim_ref.namespace)
            k8s_client.delete_persistent_volume(pv.metadata.name)
            logger.debug("Removing finalizers from the PV %s" % pv.metadata.name)
            result = k8s_client.patch_persistent_volume(pv.metadata.name, [{"op": "remove", "path": "/metadata/finalizers"}])
            logger.debug("Waiting for the PV %s to be deleted" % pv.metadata.name)
            time.sleep(1)
            try:
                while k8s_client.read_persistent_volume(pv.metadata.name) is not None:
                    logger.debug("Waiting for PVC %s in the namespace %s to be deleted"%(pv.metadata.name))
                    time.sleep(5)
            except:
                pass

            logger.debug("Recreating the PV %s" % pv.metadata.name)
            k8s_client.create_persistent_volume(pv_sanitized)
            # k8s_client.create_namespaced_persistent_volume_claim(namespace=pvc_sanitized.metadata.namespace, body=pvc_sanitized)
        else:
            logger.debug("PV %s doesn't match any migration criteria and will not be migrated" % pv.metadata.name)

    logger.debug("Migration done")

if __name__ == "__main__":

    # Default output directory path
    default_logs_dir_path = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "pv-migration-logs")

    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--context", help="Kubernetes context to use", type=str, required=True)
    parser.add_argument("-s", "--storage-class-name", help="Kubernetes storageClassName to migrate", type=str, required=True)
    parser.add_argument("-l", "--loglevel", help="Verbosity of logging (DEBUG, INFO, WARNING, ERROR, CRITICAL). [ Default: `INFO` ]", type=str, default="INFO", required=False)
    parser.add_argument("-d", "--logs-directory", help="Logs directory", type=str, default=default_logs_dir_path, required=False)
    
    args = parser.parse_args()

    main(args)